"use strict";
// Copyright (c) Mito
// Distributed under the terms of the Modified BSD License.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
/*
  Contains a wrapper around a strongly typed object that simulates a web-api.
*/
// TODO: add a class that takes a _send_ and a _receive_, and then has functions for each call of the API
// Max delay is the longest we'll wait for the API to return a value
const MAX_DELAY = 5000;
// How often we poll to see if we have a response yet
const RETRY_DELAY = 250;
const MAX_RETRIES = MAX_DELAY / RETRY_DELAY;
const getRandomId = () => {
    return '_' + Math.random().toString(36).substr(2, 9);
};
/*
  The MitoAPI class contains functions for interacting with the Mito backend.

  TODO: we should move _all_ calls to send inside of this class, and
  wrap them inside of stronger-typed functions.
*/
class MitoAPI {
    constructor(send) {
        this.send = send;
        this.unconsumedResponses = [];
    }
    /*
        The receive response function is a workaround to the fact that we _do not_ have
        a real API in practice. If/when we do have a real API, we'll get rid of this function,
        and allow the API to just make a call to a server, and wait on a response
    */
    receiveResponse(response) {
        this.unconsumedResponses.push(response);
    }
    /*
        Helper function that tries to get the response for a given ID, and returns
        the data inside the 'data' key in this response if it exists.

        Returns undefined if it does not get a response within the set timeframe
        for retries.
    */
    getResponseData(id) {
        return new Promise((resolve) => {
            let tries = 0;
            const interval = setInterval(() => {
                // Only try at most MAX_RETRIES times
                tries++;
                if (tries > MAX_RETRIES) {
                    clearInterval(interval);
                    // If we fail, we return an empty response
                    return resolve(undefined);
                }
                // See if there is an API response to this one specificially
                const index = this.unconsumedResponses.findIndex((response) => {
                    return response['id'] === id;
                });
                if (index !== -1) {
                    // Clear the interval
                    clearInterval(interval);
                    const response = this.unconsumedResponses[index];
                    this.unconsumedResponses.splice(index, 1);
                    return resolve(response['data']); // return to end execution
                }
                else {
                    console.log("Still waiting");
                }
            }, RETRY_DELAY);
        });
    }
    /*
    Returns all the CSV files in the current folder as the kernel.
  */
    getDataFiles() {
        return __awaiter(this, void 0, void 0, function* () {
            const id = getRandomId();
            this.send({
                'event': 'api_call',
                'type': 'datafiles',
                'id': id
            });
            const dataFiles = yield this.getResponseData(id);
            if (dataFiles == undefined) {
                return [];
            }
            return dataFiles;
        });
    }
    /*
        Import summaries are a mapping from step_idx -> import information for each of the
        import steps in the analysis with the given analysisName.
    */
    getImportSummary(analysisName) {
        return __awaiter(this, void 0, void 0, function* () {
            const id = getRandomId();
            this.send({
                'event': 'api_call',
                'type': 'import_summary',
                'id': id,
                'analysis_name': analysisName
            });
            const importSumary = yield this.getResponseData(id);
            if (importSumary == undefined) {
                return {};
            }
            return importSumary;
        });
    }
    /*
        Does a merge with the passed parameters, returning the ID of the edit
        event that was generated (in case you want to overwrite it).
    */
    sendMergeMessage(sheetOneIndex, sheetOneMergeKey, sheetOneSelectedColumns, sheetTwoIndex, sheetTwoMergeKey, sheetTwoSelectedColumns, 
    /*
        If you want to overwrite, you have to pass the ID of the the step that
        you want to overwrite. Not passing this argument, or passing an empty string,
        will result in no overwrite occuring (and a new stepID) being returned.
    */
    stepID) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            // If this is overwriting a merge event, then we do not need to
            // create a new id, as we already have it!
            if (stepID === undefined || stepID == '') {
                stepID = getRandomId();
            }
            (_a = window.logger) === null || _a === void 0 ? void 0 : _a.track({
                userId: window.user_id,
                event: 'button_merge_log_event',
                properties: {
                    step_id: stepID,
                    sheet_index_one: sheetOneIndex,
                    merge_key_one: sheetOneMergeKey,
                    selected_columns_one: sheetOneSelectedColumns,
                    sheet_index_two: sheetTwoIndex,
                    merge_key_two: sheetTwoMergeKey,
                    selected_columns_two: sheetTwoSelectedColumns,
                }
            });
            this.send({
                event: 'edit_event',
                type: 'merge_edit',
                step_id: stepID,
                sheet_index_one: sheetOneIndex,
                merge_key_one: sheetOneMergeKey,
                selected_columns_one: sheetOneSelectedColumns,
                sheet_index_two: sheetTwoIndex,
                merge_key_two: sheetTwoMergeKey,
                selected_columns_two: sheetTwoSelectedColumns,
            });
            return stepID;
        });
    }
    /*
        Does a pivot with the passed parameters, returning the ID of the edit
        event that was generated (in case you want to overwrite it).
    */
    sendPivotMessage(sheetIndex, pivotRows, pivotCols, values, stepID) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            // If this is overwriting a pivot event, then we do not need to
            // create a new id, as we already have it!
            if (stepID === undefined || stepID === '') {
                stepID = getRandomId();
            }
            (_a = window.logger) === null || _a === void 0 ? void 0 : _a.track({
                userId: window.user_id,
                event: 'button_pivot_log_event',
                properties: {
                    sheet_index: sheetIndex,
                    pivot_rows: pivotRows,
                    pivot_cols: pivotCols,
                    values: values,
                    step_id: stepID
                }
            });
            this.send({
                event: 'edit_event',
                type: 'pivot_edit',
                sheet_index: sheetIndex,
                pivot_rows: pivotRows,
                pivot_columns: pivotCols,
                values: values,
                step_id: stepID
            });
            return stepID;
        });
    }
}
exports.MitoAPI = MitoAPI;
//# sourceMappingURL=api.js.map